module     P3B_uux_uux_matrix
   use msamurai, only: initsamurai, exitsamurai
   use P3B_uux_uux_util, only: square
   use P3B_uux_uux_config, only: ki, &
     & include_helicity_avg_factor, include_color_avg_factor, &
     & debug_lo_diagrams, debug_nlo_diagrams, &
     & include_symmetry_factor, renormalisation, &
     & convert_to_cdr, &
     & samurai_verbosity, samurai_test, samurai_scalar
   use P3B_uux_uux_kinematics, only: &
       in_helicities, symmetry_factor, num_legs, &
       lo_qcd_couplings, corrections_are_qcd, num_light_quarks, num_gluons
   use P3B_uux_uux_model, only: Nf, NC, sqrt2, init_functions
   use P3B_uux_uux_color, only: TR, CA, CF, numcs, &
     & incolors, init_color
   use P3B_uux_uux_diagramsh0l0, only: amplitude0l0 => amplitude
   use P3B_uux_uux_amplitudeh0, only: samplitudeh0l1 => samplitude, &
        &   finite_renormalisation0 => finite_renormalisation
   use P3B_uux_uux_diagramsh1l0, only: amplitude1l0 => amplitude
   use P3B_uux_uux_amplitudeh1, only: samplitudeh1l1 => samplitude, &
        &   finite_renormalisation1 => finite_renormalisation
   use P3B_uux_uux_diagramsh2l0, only: amplitude2l0 => amplitude
   use P3B_uux_uux_amplitudeh2, only: samplitudeh2l1 => samplitude, &
        &   finite_renormalisation2 => finite_renormalisation
   use P3B_uux_uux_diagramsh3l0, only: amplitude3l0 => amplitude
   use P3B_uux_uux_amplitudeh3, only: samplitudeh3l1 => samplitude, &
        &   finite_renormalisation3 => finite_renormalisation
   use P3B_uux_uux_diagramsh4l0, only: amplitude4l0 => amplitude
   use P3B_uux_uux_amplitudeh4, only: samplitudeh4l1 => samplitude, &
        &   finite_renormalisation4 => finite_renormalisation
   use P3B_uux_uux_diagramsh5l0, only: amplitude5l0 => amplitude
   use P3B_uux_uux_amplitudeh5, only: samplitudeh5l1 => samplitude, &
        &   finite_renormalisation5 => finite_renormalisation
   use P3B_uux_uux_dipoles, only: insertion_operator

   implicit none
   save

   private

   public :: initgolem, exitgolem, samplitude
   public :: samplitudel0, samplitudel1
   public :: ir_subtraction, color_correlated_lo2, spin_correlated_lo2
contains
   !---#[ subroutine initgolem :
   subroutine     initgolem(is_first)
      implicit none
      logical, optional, intent(in) :: is_first

      logical :: init_third_party

      if(present(is_first)) then
         init_third_party = is_first
      else
         init_third_party = .true.
      end if
      if (init_third_party) then
         call initsamurai('diag',samurai_scalar,&
         &                samurai_verbosity,samurai_test)
      end if
      call init_functions()
      call init_color()
   end subroutine initgolem
   !---#] subroutine initgolem :
   !---#[ subroutine exitgolem :
   subroutine     exitgolem(is_last)
      use P3B_uux_uux_groups, only: tear_down_golem95
      implicit none
      logical, optional, intent(in) :: is_last

      logical :: exit_third_party

      if(present(is_last)) then
         exit_third_party = is_last
      else
         exit_third_party = .true.
      end if
      if (exit_third_party) then
         call exitsamurai()
         call tear_down_golem95()

      end if
   end subroutine exitgolem
   !---#] subroutine exitgolem :
   !---#[ subroutine samplitude :
   subroutine     samplitude(vecs, scale2, amp, ok, h)
      use P3B_uux_uux_config, only: &
         & debug_lo_diagrams, debug_nlo_diagrams, logfile
      use P3B_uux_uux_kinematics, only: &
         & inspect_kinematics, init_event
      implicit none
      real(ki), dimension(8, 4), intent(in) :: vecs
      double precision, intent(in) :: scale2
      double precision, dimension(4), intent(out) :: amp
      logical, intent(out), optional :: ok
      integer, intent(in), optional :: h

      logical :: my_ok

      ! used for m=0 QCD renormalisation
      real(ki) :: beta0

      call init_event(vecs)

      if(debug_lo_diagrams .or. debug_nlo_diagrams) then
         write(logfile,'(A7)') "<event>"
         call inspect_kinematics(logfile)
      end if

      
      if (present(h)) then
         amp(1) = samplitudel0(vecs, h)
      else
         amp(1)   = samplitudel0(vecs)
      end if
      if (present(h)) then
         amp((/4,3,2/)) = samplitudel1(vecs, scale2, my_ok, h)
      else
         amp((/4,3,2/)) = samplitudel1(vecs, scale2, my_ok)
      end if
      select case (renormalisation)
      case (0)
         ! no renormalisation
      case (1)
         if(corrections_are_qcd) then
            beta0 = (11.0_ki * CA - 4.0_ki * TR * NF) / 6.0_ki
            amp(3) = amp(3) - lo_qcd_couplings * beta0 * amp(1)
            amp(2) = amp(2) + lo_qcd_couplings * CA / 6.0_ki * amp(1)
         end if
      case default
         ! not implemented
         print*, "In P3B_uux_uux_matrix:"
         print*, "  invalid value for renormalisation=", renormalisation
         stop
      end select
      if (convert_to_cdr) then
         ! Scheme conversion for infrared structure
         ! Reference:
         ! S. Catani, M. H. Seymour, Z. Trocsanyi,
         ! ``Regularisation scheme independence and unitarity
         !   in QCD cross-sections,''
         ! Phys.Rev. D 55 (1997) 6819
         ! arXiv:hep-ph/9610553
         amp(2) = amp(2) - amp(1) * (&
           &          num_light_quarks * 0.5_ki * CF &
           &        + num_gluons * 1.0_ki/6.0_ki * CA)
      end if
      if (present(ok)) ok = my_ok

      if(debug_lo_diagrams .or. debug_nlo_diagrams) then
         write(logfile,'(A25,E24.16,A3)') &
            & "<result kind='lo' value='", amp(1), "'/>"
         write(logfile,'(A33,E24.16,A3)') &
            & "<result kind='nlo-finite' value='", amp(2), "'/>"
         write(logfile,'(A33,E24.16,A3)') &
            & "<result kind='nlo-single' value='", amp(3), "'/>"
         write(logfile,'(A33,E24.16,A3)') &
            & "<result kind='nlo-double' value='", amp(4), "'/>"
         if(my_ok) then
            write(logfile,'(A30)') "<flag name='ok' status='yes'/>"
         else
            write(logfile,'(A29)') "<flag name='ok' status='no'/>"
         end if
         write(logfile,'(A8)') "</event>"
      end if
   end subroutine samplitude
   !---#] subroutine samplitude :
   !---#[ function samplitudel0 :
   function     samplitudel0(vecs, h) result(amp)
      use P3B_uux_uux_config, only: logfile
      implicit none
      real(ki), dimension(8, 4), intent(in) :: vecs
      integer, optional, intent(in) :: h
      real(ki) :: amp, heli_amp
      complex(ki), dimension(numcs) :: color_vector
      logical, dimension(0:255) :: eval_heli

      if (present(h)) then
         eval_heli(:) = .false.
         eval_heli(h) = .true.
      else
         eval_heli(:) = .true.
      end if

      amp = 0.0_ki
      if (eval_heli(0)) then
         if (debug_lo_diagrams) then
            write(logfile,*) "<helicity index='0' >"
         end if
         color_vector = amplitude0l0()
         heli_amp = square(color_vector)
         if (debug_lo_diagrams) then
            write(logfile,'(A25,E24.16,A3)') &
                & "<result kind='lo' value='", heli_amp, "'/>"
            write(logfile,*) "</helicity>"
         end if
         amp = amp + heli_amp
      end if
      if (eval_heli(1)) then
         if (debug_lo_diagrams) then
            write(logfile,*) "<helicity index='1' >"
         end if
         color_vector = amplitude1l0()
         heli_amp = square(color_vector)
         if (debug_lo_diagrams) then
            write(logfile,'(A25,E24.16,A3)') &
                & "<result kind='lo' value='", heli_amp, "'/>"
            write(logfile,*) "</helicity>"
         end if
         amp = amp + heli_amp
      end if
      if (eval_heli(2)) then
         if (debug_lo_diagrams) then
            write(logfile,*) "<helicity index='2' >"
         end if
         color_vector = amplitude2l0()
         heli_amp = square(color_vector)
         if (debug_lo_diagrams) then
            write(logfile,'(A25,E24.16,A3)') &
                & "<result kind='lo' value='", heli_amp, "'/>"
            write(logfile,*) "</helicity>"
         end if
         amp = amp + heli_amp
      end if
      if (eval_heli(3)) then
         if (debug_lo_diagrams) then
            write(logfile,*) "<helicity index='3' >"
         end if
         color_vector = amplitude3l0()
         heli_amp = square(color_vector)
         if (debug_lo_diagrams) then
            write(logfile,'(A25,E24.16,A3)') &
                & "<result kind='lo' value='", heli_amp, "'/>"
            write(logfile,*) "</helicity>"
         end if
         amp = amp + heli_amp
      end if
      if (eval_heli(4)) then
         if (debug_lo_diagrams) then
            write(logfile,*) "<helicity index='4' >"
         end if
         color_vector = amplitude4l0()
         heli_amp = square(color_vector)
         if (debug_lo_diagrams) then
            write(logfile,'(A25,E24.16,A3)') &
                & "<result kind='lo' value='", heli_amp, "'/>"
            write(logfile,*) "</helicity>"
         end if
         amp = amp + heli_amp
      end if
      if (eval_heli(5)) then
         if (debug_lo_diagrams) then
            write(logfile,*) "<helicity index='5' >"
         end if
         color_vector = amplitude5l0()
         heli_amp = square(color_vector)
         if (debug_lo_diagrams) then
            write(logfile,'(A25,E24.16,A3)') &
                & "<result kind='lo' value='", heli_amp, "'/>"
            write(logfile,*) "</helicity>"
         end if
         amp = amp + heli_amp
      end if
      if (include_helicity_avg_factor) then
         amp = amp / real(in_helicities, ki)
      end if
      if (include_color_avg_factor) then
         amp = amp / incolors
      end if
      if (include_symmetry_factor) then
         amp = amp / real(symmetry_factor, ki)
      end if
   end function samplitudel0
   !---#] function samplitudel0 :
   !---#[ function samplitudel1 :
   function     samplitudel1(vecs,scale2,ok,h) result(amp)
      use P3B_uux_uux_config, only: &
         & debug_nlo_diagrams, logfile
      implicit none
      real(ki), dimension(8, 4), intent(in) :: vecs
      logical, intent(out) :: ok
      double precision, intent(in) :: scale2
      integer, optional, intent(in) :: h
      real(ki), dimension(-2:0) :: amp, heli_amp
      logical :: my_ok
      logical, dimension(0:255) :: eval_heli
      real(ki) :: fr

      if (present(h)) then
         eval_heli(:) = .false.
         eval_heli(h) = .true.
      else
         eval_heli(:) = .true.
      end if

      amp(:) = 0.0_ki
      ok = .true.
      if (eval_heli(0)) then
         if(debug_nlo_diagrams) then
            write(logfile,*) "<helicity index='0'>"
         end if
         heli_amp = samplitudeh0l1(real(scale2,ki),my_ok)
         if (corrections_are_qcd) then
            fr = finite_renormalisation0(real(scale2,ki))
            heli_amp(0) = heli_amp(0) + fr
         end if
         ok = ok .and. my_ok
         amp = amp + heli_amp

         if(debug_nlo_diagrams) then
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-finite' value='", heli_amp(0), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-single' value='", heli_amp(-1), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-double' value='", heli_amp(-2), "'/>"
            if (corrections_are_qcd) then
               write(logfile,'(A30,E24.16,A3)') &
                   & "<result kind='fin-ren' value='", heli_amp(-2), "'/>"
            end if
            if(my_ok) then
               write(logfile,'(A30)') "<flag name='ok' status='yes'/>"
            else
               write(logfile,'(A29)') "<flag name='ok' status='no'/>"
            end if
            write(logfile,*) "</helicity>"
         end if
      end if
      if (eval_heli(1)) then
         if(debug_nlo_diagrams) then
            write(logfile,*) "<helicity index='1'>"
         end if
         heli_amp = samplitudeh1l1(real(scale2,ki),my_ok)
         if (corrections_are_qcd) then
            fr = finite_renormalisation1(real(scale2,ki))
            heli_amp(0) = heli_amp(0) + fr
         end if
         ok = ok .and. my_ok
         amp = amp + heli_amp

         if(debug_nlo_diagrams) then
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-finite' value='", heli_amp(0), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-single' value='", heli_amp(-1), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-double' value='", heli_amp(-2), "'/>"
            if (corrections_are_qcd) then
               write(logfile,'(A30,E24.16,A3)') &
                   & "<result kind='fin-ren' value='", heli_amp(-2), "'/>"
            end if
            if(my_ok) then
               write(logfile,'(A30)') "<flag name='ok' status='yes'/>"
            else
               write(logfile,'(A29)') "<flag name='ok' status='no'/>"
            end if
            write(logfile,*) "</helicity>"
         end if
      end if
      if (eval_heli(2)) then
         if(debug_nlo_diagrams) then
            write(logfile,*) "<helicity index='2'>"
         end if
         heli_amp = samplitudeh2l1(real(scale2,ki),my_ok)
         if (corrections_are_qcd) then
            fr = finite_renormalisation2(real(scale2,ki))
            heli_amp(0) = heli_amp(0) + fr
         end if
         ok = ok .and. my_ok
         amp = amp + heli_amp

         if(debug_nlo_diagrams) then
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-finite' value='", heli_amp(0), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-single' value='", heli_amp(-1), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-double' value='", heli_amp(-2), "'/>"
            if (corrections_are_qcd) then
               write(logfile,'(A30,E24.16,A3)') &
                   & "<result kind='fin-ren' value='", heli_amp(-2), "'/>"
            end if
            if(my_ok) then
               write(logfile,'(A30)') "<flag name='ok' status='yes'/>"
            else
               write(logfile,'(A29)') "<flag name='ok' status='no'/>"
            end if
            write(logfile,*) "</helicity>"
         end if
      end if
      if (eval_heli(3)) then
         if(debug_nlo_diagrams) then
            write(logfile,*) "<helicity index='3'>"
         end if
         heli_amp = samplitudeh3l1(real(scale2,ki),my_ok)
         if (corrections_are_qcd) then
            fr = finite_renormalisation3(real(scale2,ki))
            heli_amp(0) = heli_amp(0) + fr
         end if
         ok = ok .and. my_ok
         amp = amp + heli_amp

         if(debug_nlo_diagrams) then
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-finite' value='", heli_amp(0), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-single' value='", heli_amp(-1), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-double' value='", heli_amp(-2), "'/>"
            if (corrections_are_qcd) then
               write(logfile,'(A30,E24.16,A3)') &
                   & "<result kind='fin-ren' value='", heli_amp(-2), "'/>"
            end if
            if(my_ok) then
               write(logfile,'(A30)') "<flag name='ok' status='yes'/>"
            else
               write(logfile,'(A29)') "<flag name='ok' status='no'/>"
            end if
            write(logfile,*) "</helicity>"
         end if
      end if
      if (eval_heli(4)) then
         if(debug_nlo_diagrams) then
            write(logfile,*) "<helicity index='4'>"
         end if
         heli_amp = samplitudeh4l1(real(scale2,ki),my_ok)
         if (corrections_are_qcd) then
            fr = finite_renormalisation4(real(scale2,ki))
            heli_amp(0) = heli_amp(0) + fr
         end if
         ok = ok .and. my_ok
         amp = amp + heli_amp

         if(debug_nlo_diagrams) then
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-finite' value='", heli_amp(0), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-single' value='", heli_amp(-1), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-double' value='", heli_amp(-2), "'/>"
            if (corrections_are_qcd) then
               write(logfile,'(A30,E24.16,A3)') &
                   & "<result kind='fin-ren' value='", heli_amp(-2), "'/>"
            end if
            if(my_ok) then
               write(logfile,'(A30)') "<flag name='ok' status='yes'/>"
            else
               write(logfile,'(A29)') "<flag name='ok' status='no'/>"
            end if
            write(logfile,*) "</helicity>"
         end if
      end if
      if (eval_heli(5)) then
         if(debug_nlo_diagrams) then
            write(logfile,*) "<helicity index='5'>"
         end if
         heli_amp = samplitudeh5l1(real(scale2,ki),my_ok)
         if (corrections_are_qcd) then
            fr = finite_renormalisation5(real(scale2,ki))
            heli_amp(0) = heli_amp(0) + fr
         end if
         ok = ok .and. my_ok
         amp = amp + heli_amp

         if(debug_nlo_diagrams) then
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-finite' value='", heli_amp(0), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-single' value='", heli_amp(-1), "'/>"
            write(logfile,'(A33,E24.16,A3)') &
                & "<result kind='nlo-double' value='", heli_amp(-2), "'/>"
            if (corrections_are_qcd) then
               write(logfile,'(A30,E24.16,A3)') &
                   & "<result kind='fin-ren' value='", heli_amp(-2), "'/>"
            end if
            if(my_ok) then
               write(logfile,'(A30)') "<flag name='ok' status='yes'/>"
            else
               write(logfile,'(A29)') "<flag name='ok' status='no'/>"
            end if
            write(logfile,*) "</helicity>"
         end if
      end if
      if (include_helicity_avg_factor) then
         amp = amp / real(in_helicities, ki)
      end if
      if (include_color_avg_factor) then
         amp = amp / incolors
      end if
      if (include_symmetry_factor) then
         amp = amp / real(symmetry_factor, ki)
      end if
   end function samplitudel1
   !---#] function samplitudel1 :
   !---#[ subroutine ir_subtraction :
   subroutine     ir_subtraction(vecs,scale2,amp)
      use P3B_uux_uux_kinematics, only: &
         & init_event
      implicit none
      real(ki), dimension(8, 4), intent(in) :: vecs
      double precision, intent(in) :: scale2
      real(ki), dimension(2), intent(out) :: amp
      real(ki), dimension(2) :: heli_amp
      complex(ki), dimension(numcs,numcs,2) :: oper
      complex(ki), dimension(numcs) :: color_vectorl0

      call init_event(vecs)

      oper = insertion_operator(real(scale2,ki), vecs)
      amp(:) = 0.0_ki
      color_vectorl0 = amplitude0l0()
      heli_amp(1) = square(color_vectorl0, oper(:,:,1))
      heli_amp(2) = square(color_vectorl0, oper(:,:,2))
      amp = amp + heli_amp
      color_vectorl0 = amplitude1l0()
      heli_amp(1) = square(color_vectorl0, oper(:,:,1))
      heli_amp(2) = square(color_vectorl0, oper(:,:,2))
      amp = amp + heli_amp
      color_vectorl0 = amplitude2l0()
      heli_amp(1) = square(color_vectorl0, oper(:,:,1))
      heli_amp(2) = square(color_vectorl0, oper(:,:,2))
      amp = amp + heli_amp
      color_vectorl0 = amplitude3l0()
      heli_amp(1) = square(color_vectorl0, oper(:,:,1))
      heli_amp(2) = square(color_vectorl0, oper(:,:,2))
      amp = amp + heli_amp
      color_vectorl0 = amplitude4l0()
      heli_amp(1) = square(color_vectorl0, oper(:,:,1))
      heli_amp(2) = square(color_vectorl0, oper(:,:,2))
      amp = amp + heli_amp
      color_vectorl0 = amplitude5l0()
      heli_amp(1) = square(color_vectorl0, oper(:,:,1))
      heli_amp(2) = square(color_vectorl0, oper(:,:,2))
      amp = amp + heli_amp
      if (include_helicity_avg_factor) then
         amp = amp / real(in_helicities, ki)
      end if
      if (include_color_avg_factor) then
         amp = amp / incolors
      end if
      if (include_symmetry_factor) then
         amp = amp / real(symmetry_factor, ki)
      end if
   end subroutine ir_subtraction
   !---#] subroutine ir_subtraction :
   !---#[ color correlated ME :
   pure subroutine color_correlated_lo(color_vector,res)
      use P3B_uux_uux_color, only: T1T1, &
      & T1T2, &
      & T1T3, &
      & T1T4, &
      & T2T2, &
      & T2T3, &
      & T2T4, &
      & T3T3, &
      & T3T4, &
      & T4T4
      implicit none
      complex(ki), dimension(numcs), intent(in) :: color_vector
      real(ki), dimension(num_legs,num_legs), intent(out) :: res
      res(:,:)=0.0_ki
      res(1,1) = square(color_vector,T1T1)
      res(1,1) = res(1,1)
      res(1,2) = square(color_vector,T1T2)
      res(2,1) = res(1,2)
      res(1,3) = square(color_vector,T1T3)
      res(3,1) = res(1,3)
      res(1,4) = square(color_vector,T1T4)
      res(4,1) = res(1,4)
      res(2,2) = square(color_vector,T2T2)
      res(2,2) = res(2,2)
      res(2,3) = square(color_vector,T2T3)
      res(3,2) = res(2,3)
      res(2,4) = square(color_vector,T2T4)
      res(4,2) = res(2,4)
      res(3,3) = square(color_vector,T3T3)
      res(3,3) = res(3,3)
      res(3,4) = square(color_vector,T3T4)
      res(4,3) = res(3,4)
      res(4,4) = square(color_vector,T4T4)
      res(4,4) = res(4,4)
   end subroutine color_correlated_lo

   subroutine     color_correlated_lo2(vecs,borncc)
      implicit none
      real(ki), dimension(num_legs, 4), intent(in) :: vecs
      real(ki), dimension(num_legs,num_legs), intent(out) :: borncc
      real(ki), dimension(num_legs,num_legs) :: borncc_heli
      complex(ki), dimension(numcs) :: color_vector

      borncc(:,:) = 0.0_ki
      color_vector = amplitude0l0()
      call color_correlated_lo(color_vector,borncc_heli)
      ! The minus is part in the definition according to PowHEG Box.
      ! Since they use it we include it:
      borncc(:,:) = borncc(:,:) - borncc_heli(:,:)
      color_vector = amplitude1l0()
      call color_correlated_lo(color_vector,borncc_heli)
      borncc(:,:) = borncc(:,:) - borncc_heli(:,:)
      color_vector = amplitude2l0()
      call color_correlated_lo(color_vector,borncc_heli)
      borncc(:,:) = borncc(:,:) - borncc_heli(:,:)
      color_vector = amplitude3l0()
      call color_correlated_lo(color_vector,borncc_heli)
      borncc(:,:) = borncc(:,:) - borncc_heli(:,:)
      color_vector = amplitude4l0()
      call color_correlated_lo(color_vector,borncc_heli)
      borncc(:,:) = borncc(:,:) - borncc_heli(:,:)
      color_vector = amplitude5l0()
      call color_correlated_lo(color_vector,borncc_heli)
      borncc(:,:) = borncc(:,:) - borncc_heli(:,:)
      if (include_helicity_avg_factor) then
         borncc = borncc / real(in_helicities, ki)
      end if
      if (include_color_avg_factor) then
         borncc = borncc / incolors
      end if
      if (include_symmetry_factor) then
         borncc = borncc / real(symmetry_factor, ki)
      end if
   end subroutine color_correlated_lo2
   !---#] color correlated ME :
   !---#[ spin correlated ME :
   subroutine spin_correlated_lo2(vecs, bornsc)
      use P3B_uux_uux_kinematics
      implicit none
      real(ki), dimension(num_legs, 4), intent(in) :: vecs
      real(ki), dimension(num_legs,4,4) :: bornsc
      complex(ki), dimension(4,4) :: tens
      complex(ki) :: pp, pm, mp, mm

      bornsc(:,:,:) = 0.0_ki
      !---#[ Initialize helicity amplitudes :
      !---#] Initialize helicity amplitudes :
      !---#[ Initialize polarization vectors :
      !---#] Initialize polarization vectors :
      ! Note: By omitting the imaginary parts we lose a term:
      !   Imag(B_j(mu,nu)) = i_ * e_(k_j, mu, q_j, nu) * |Born|^2
      ! where q_j is the reference momentum chosen for the paticle
      ! of momentum k_j. This term should, however not be phenomenologically
      ! relevant.
      if (include_helicity_avg_factor) then
         bornsc = bornsc / real(in_helicities, ki)
      end if
      if (include_color_avg_factor) then
         bornsc = bornsc / incolors
      end if
      if (include_symmetry_factor) then
         bornsc = bornsc / real(symmetry_factor, ki)
      end if
   end subroutine spin_correlated_lo2
   !---#] spin correlated ME :
   !---#[ construct polarisation tensor :
   pure subroutine construct_polarization_tensor(eps1, eps2, tens)
      implicit none
      complex(ki), dimension(0:3), intent(in) :: eps1, eps2
      complex(ki), dimension(0:3,0:3), intent(out) :: tens

      integer :: mu, nu

      do mu = 0,3
         do nu = 0, 3
            tens(mu,nu) = eps1(mu) * eps2(nu)
         end do
      end do
   end  subroutine construct_polarization_tensor
   !---#] construct polarisation tensor :
   pure function square_0l_0l_sc(color_vector1, color_vector2) result(amp)
      use P3B_uux_uux_color, only: cmat => CC
      implicit none
      complex(ki), dimension(numcs), intent(in) :: color_vector1, color_vector2
      complex(ki) :: amp
      complex(ki), dimension(numcs) :: v1, v2

      v1 = matmul(cmat, color_vector2)
      v2 = conjg(color_vector1)
      amp = sum(v1(:) * v2(:))
   end function  square_0l_0l_sc

end module P3B_uux_uux_matrix
