program test
   use P3B_uux_uux_config, only: ki, logfile
   use P3B_uux_uux_kinematics, only: dotproduct
   use P3B_uux_uux_model, only: parse
   use P3B_uux_uux_matrix, only: samplitude, &
     & initgolem, ir_subtraction
   use P3B_uux_uux_color, only: numcs, CA
   use P3B_uux_uux_rambo, only: ramb

   implicit none

   integer, parameter :: NEVT = 1

   integer :: ievt, ierr
   real(ki), dimension(8, 4) :: vecs
   double precision :: scale2
   double precision, dimension(0:3) :: amp

   open(unit=logfile,status='unknown',action='write',file='debug.xml')

   open(unit=10,status='old',action='read',file='param.dat',iostat=ierr)
   if(ierr .eq. 0) then
      call parse(10)
      close(unit=10)
   else
      print*, "No file 'param.dat' found. Using defaults"
   end if

   call initgolem()

   call random_seed

   do ievt = 1, NEVT
      call ramb(1.0E+00_ki**2, vecs)

      scale2 = 2.0_ki * dotproduct(vecs(1,:), vecs(2,:))

      call samplitude(vecs, scale2, amp)
      write(*,'(A1,1x,A15,1x,G23.16)') "#", "LO:", amp(0)
      write(*,'(A1,1x,A15,1x,G23.16)') "#", "NLO, finite part:", &
          &    amp(1)/amp(0)
      write(*,'(A1,1x,A15,1x,G23.16)') "#", "NLO, single pole:", &
          &    amp(2)/amp(0)
      write(*,'(A1,1x,A15,1x,G23.16)') "#", "NLO, double pole:", &
          &    amp(3)/amp(0)
      
      call ir_subtraction(vecs, scale2, amp(2:3))
      write(*,'(A1,1x,A15,1x,G23.16)') "#", "IR,  single pole:", &
         & amp(2)/amp(0)
      write(*,'(A1,1x,A15,1x,G23.16)') "#", "IR,  double pole:", &
         & amp(3)/amp(0)
      
   end do

   close(logfile)
end program test
